/*
 * Copyright 2016 - 2022 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#include <ox/claw/write.hpp>

#include <nostalgia/core/media.hpp>

#include "tilesheeteditormodel.hpp"

namespace nostalgia::core {

TileSheetEditorModel::TileSheetEditorModel(Context *ctx, const ox::String &path) {
	m_ctx = ctx;
	m_path = path;
	oxRequireT(img, readObj<TileSheet>(ctx, path.c_str()));
	m_img = *img;
	oxThrowError(readObj<Palette>(ctx, m_img.defaultPalette).moveTo(&m_pal));
	m_undoStack.changeTriggered.connect(this, &TileSheetEditorModel::markUpdated);
}

void TileSheetEditorModel::cut() {
}

void TileSheetEditorModel::copy() {
}

void TileSheetEditorModel::paste() {
}

void TileSheetEditorModel::drawCommand(const geo::Point &pt, std::size_t palIdx) noexcept {
	auto &activeSubSheet = m_img.getSubSheet(m_activeSubsSheetIdx);
	const auto idx = ptToIdx(pt, activeSubSheet.columns);
	if (idx >= activeSubSheet.pixelCnt(m_img.bpp)) {
		return;
	}
	if (m_ongoingDrawCommand) {
		m_updated = m_updated || m_ongoingDrawCommand->append(idx);
	} else {
		if (activeSubSheet.getPixel(m_img.bpp, idx) != palIdx) {
			pushCommand(new DrawCommand(&m_img, m_activeSubsSheetIdx, idx, palIdx));
		}
	}
}

void TileSheetEditorModel::endDrawCommand() noexcept {
	m_ongoingDrawCommand = nullptr;
}

void TileSheetEditorModel::addSubsheet(const TileSheet::SubSheetIdx &parentIdx) noexcept {
	pushCommand(new AddSubSheetCommand(&m_img, parentIdx));
}

void TileSheetEditorModel::rmSubsheet(const TileSheet::SubSheetIdx &idx) noexcept {
	pushCommand(new RmSubSheetCommand(&m_img, idx));
}

ox::Error TileSheetEditorModel::updateSubsheet(const TileSheet::SubSheetIdx &idx, const ox::String &name, int cols, int rows) noexcept {
	pushCommand(new UpdateSubSheetCommand(&m_img, idx, name, cols, rows));
	return OxError(0);
}

void TileSheetEditorModel::setActiveSubsheet(const TileSheet::SubSheetIdx &idx) noexcept {
	m_activeSubsSheetIdx = idx;
	this->activeSubsheetChanged.emit(m_activeSubsSheetIdx);
}

void TileSheetEditorModel::select(const geo::Point &pt) noexcept {
	if (!m_selectionOngoing) {
		m_selectionPt1 = pt;
		m_selectionOngoing = true;
	}
	if (m_selectionPt2 != pt) {
		m_selectionPt2 = pt;
		m_selectionBounds = {m_selectionPt1, m_selectionPt2};
		m_updated = true;
	}
}

void TileSheetEditorModel::completeSelection() noexcept {
	m_selectionOngoing = false;
}

void TileSheetEditorModel::clearSelection() noexcept {
	m_updated = true;
	m_selectionPt1 = {-1, -1};
	m_selectionPt2 = {-1, -1};
	m_selectionBounds = {m_selectionPt1, m_selectionPt2};
}

bool TileSheetEditorModel::updated() const noexcept {
	return m_updated;
}

ox::Error TileSheetEditorModel::markUpdated(int cmdId) noexcept {
	m_updated = true;
	switch (static_cast<CommandId>(cmdId)) {
		case CommandId::AddSubSheet:
		case CommandId::RmSubSheet: {
			// make sure the current active SubSheet is still valid
			auto idx = m_img.validateSubSheetIdx(m_activeSubsSheetIdx);
			if (idx != m_activeSubsSheetIdx) {
				setActiveSubsheet(idx);
			}
			break;
		}
		case CommandId::Draw:
		case CommandId::UpdateSubSheet:
			break;
	}
	return OxError(0);
}

void TileSheetEditorModel::ackUpdate() noexcept {
	m_updated = false;
}

ox::Error TileSheetEditorModel::saveFile() noexcept {
	auto sctx = applicationData<studio::StudioContext>(m_ctx);
	oxReturnError(sctx->project->writeObj(m_path, &m_img));
	return m_ctx->assetManager.setAsset(m_path, m_img).error;
}

bool TileSheetEditorModel::pixelSelected(std::size_t idx) const noexcept {
	auto s = activeSubSheet();
	auto pt = idxToPt(idx, s->columns);
	return m_selectionBounds.contains(pt);
}

void TileSheetEditorModel::getFillPixels(bool *pixels, const geo::Point &pt, int oldColor) const noexcept {
	auto &activeSubSheet = *this->activeSubSheet();
	const auto tileIdx = [activeSubSheet](const geo::Point &pt) noexcept {
		return ptToIdx(pt, activeSubSheet.columns) / PixelsPerTile;
	};
	// get points
	const auto leftPt = pt + geo::Point(-1, 0);
	const auto rightPt = pt + geo::Point(1, 0);
	const auto topPt = pt + geo::Point(0, -1);
	const auto bottomPt = pt + geo::Point(0, 1);
	// calculate indices
	const auto idx = ptToIdx(pt, activeSubSheet.columns);
	const auto leftIdx = ptToIdx(leftPt, activeSubSheet.columns);
	const auto rightIdx = ptToIdx(rightPt, activeSubSheet.columns);
	const auto topIdx = ptToIdx(topPt, activeSubSheet.columns);
	const auto bottomIdx = ptToIdx(bottomPt, activeSubSheet.columns);
	const auto tile = tileIdx(pt);
	// mark pixels to update
	pixels[idx % PixelsPerTile] = true;
	if (!pixels[leftIdx % PixelsPerTile] && tile == tileIdx(leftPt) && activeSubSheet.pixels[leftIdx] == oldColor) {
		getFillPixels(pixels, leftPt, oldColor);
	}
	if (!pixels[rightIdx % PixelsPerTile] && tile == tileIdx(rightPt) && activeSubSheet.pixels[rightIdx] == oldColor) {
		getFillPixels(pixels, rightPt, oldColor);
	}
	if (!pixels[topIdx % PixelsPerTile] && tile == tileIdx(topPt) && activeSubSheet.pixels[topIdx] == oldColor) {
		getFillPixels(pixels, topPt, oldColor);
	}
	if (!pixels[bottomIdx % PixelsPerTile] && tile == tileIdx(bottomPt) && activeSubSheet.pixels[bottomIdx] == oldColor) {
		getFillPixels(pixels, bottomPt, oldColor);
	}
}

void TileSheetEditorModel::pushCommand(studio::UndoCommand *cmd) noexcept {
	m_undoStack.push(cmd);
	m_ongoingDrawCommand = dynamic_cast<DrawCommand*>(cmd);
	m_updated = true;
}

}
